PORT=1789
HOST='localhost'
REL_IS_LOCAL_FOR='is_local_for'
REL_IS_BOOST_OF='is_boost_of'
REL_IS_BOOST_OF='is_boost_of'
REL_IS_REPLY_TO='is_reply_to'
REL_HAS_ID = 'has_id'
REL_TO_ID = 'to_id'
REL_READ_STATUS='rel_is_read'
NODE_READ_STATUS_IS_READ='read'
NODE_READ_STATUS_IS_UNREAD='unread'
REL_GRAPHTIME='when'


BLUR_MINUTES=5   # 1 minutes grace period / blur

LIMNODES=20
SCORE_TYPE = 'All'
LIM_TIMELINE=20
LIM_CONVO=3


import os,sys; sys.path.insert(0,'..')
FORK_AVAILABLE=False
os.environ['FORK_AVAILABLE']=str(FORK_AVAILABLE)
path_imports = path_self = os.path.realpath(__file__)
path_gui = os.path.abspath(os.path.join(path_imports,'..','gui'))
path_code = os.path.abspath(os.path.join(path_imports,'..','..'))
path_web = path_gui
path_static = os.path.join(path_web,'static')
path_templates = os.path.join(path_web,'templates')

from pprint import pprint, pformat
import textwrap,random
import numpy as np
import click
from functools import cached_property, lru_cache, total_ordering
cache = lru_cache(maxsize=None)
from typing import Optional,Union,Dict,List
import numpy as np
from mastodon.errors import MastodonNotFoundError, MastodonNetworkError
from mastodon.utility import AttribAccessDict
from mastodon import Mastodon, AttribAccessDict
from datetime import datetime, timedelta, timezone
import datetime as dt
from scipy.stats import gmean
from math import sqrt
import json
from pprint import pprint
import pandas as pd
from IPython.display import Markdown, display
def printm(x): display(Markdown(x))
import webbrowser,json, re
from collections import UserDict,UserList,defaultdict
def unhtml(x): 
    x = x.replace('<br>','\n')
    return re.sub(r'<.*?>', '', x).strip()
import networkx as nx
import pandas as pd
import warnings
from sqlitedict import SqliteDict
warnings.filterwarnings('ignore')
import requests
from itertools import islice
import logging
import logging as log
logging.getLogger().setLevel(logging.INFO)  # choose your level here
from tqdm import tqdm
import mastodon
from multiprocessing.pool import ThreadPool




path_data = os.path.expanduser('~/.mastotron')
path_srvr = os.path.join(path_data, 'db.webview')
path_env = os.path.join(path_data, 'config.json')
path_db = os.path.join(path_data, 'db.sqlitedict')
path_tinydb = os.path.join(path_data, 'db.tinydb')
path_blitzdb = os.path.join(path_data, 'db.blitzdb')
path_cogdb = os.path.join(path_data, 'db.codgb')
app_name = 'mastotron'
API = None

from .utils import *
from .mastotron import *
from .post import *
from .poster import *
from .postlist import *
from .postnet import *
from .db import *
from .graphdb import *
from .htmlfmt import *



from .imports import *

class PostList(UserList):
    def __init__(self, data_iter=[], lim=None):
        from .post import Post
        # def iterr():
        #     seen=set()
        #     for post in data_iter:
        #         post = Post(post)
        #         if post:
        #             post = post.source
        #             if post not in seen:
        #                 yield post
        #                 seen.add(post)
        def iterr(): 
            yield from (Post(x) for x in data_iter)
        l=list(islice(iterr(), lim))
        super().__init__(l)
        self.sort_chron()

    def __str__(self):
        return str([p._id for p in self])

    def __hash__(self):
        return hash('____'.join(p._id for p in self))
    def __add__(self, other):
        def iterr():
            yield from self
            yield from other
        return PostList(iterr())
    
    def __sub__(self, other):
        bad_ids = set(other)
        def iterr():
            for x in self:
                if x not in bad_ids:
                    yield x
        return PostList(iterr())


    
    def sort_chron(self):
        self.sort(key=lambda post: post.timestamp if post.timestamp else 0, reverse=False)

    def sort_score(self, score_type=SCORE_TYPE):
        self.sort(key=lambda post: post.score(score_type), reverse=True)
    
    def network(self):
        from .postnet import PostNet
        return PostNet(self)
    @cached_property
    def net(self): return self.network()
    
    def graph(self, *args, **kwargs): return self.net.graph(*args,**kwargs)
    @property
    def g(self): return self.graph()
    
                
    @property
    def posters(self):
        return {post.author for post in self}

    def interrelate(self):
        context_d = dict((post._data.get('id'), post) for post in self)
        for post in self:
            if post.is_reply and not post.out(REL_IS_REPLY_TO):
                post2 = context_d.get(post.in_reply_to_id)
                if post2:
                    post.relate(post2, rel=REL_IS_REPLY_TO)
                    print(post,'-->',post2)


from .imports import *
gdb = None
class TronDB:
    @property
    def gdb(self):
        global gdb
        if gdb is None:
            from cog.torque import Graph
            # try:
            gdb = Graph(
                'cogdb', 
                cog_home=os.path.basename(path_cogdb), 
                cog_path_prefix=os.path.dirname(path_cogdb),
                enable_caching=False
            )
            # except Exception as e:
                # # print(f'!! {e} !!')
                # rmfile(path_cogdb)
                # gdb = Graph(
                #     'cogdb', 
                #     cog_home=os.path.basename(path_cogdb), 
                #     cog_path_prefix=os.path.dirname(path_cogdb)
                # )

        return gdb
    
    def relate(self, obj1, obj2, rel):
        id1=str(obj1)
        id2=str(obj2)
        rel=str(rel)
        if id1 and id2 and rel:
            self.gdb.put(id1,rel,id2)
    
    def unrelate(self, obj1, obj2, rel):
        id1=str(obj1)
        id2=str(obj2)
        rel=str(rel)
        if id1 and id2 and rel:
            self.gdb.drop(id1,rel,id2)

    def get_rels_out(self, obj, rel):
        return [d.get('id') for d in self.gdb.v(str(obj)).out(rel).all().get('result') if d.get('id')]
    def get_rels_inc(self, obj, rel):
        return [d.get('id') for d in self.gdb.v(str(obj)).inc(rel).all().get('result') if d.get('id')]
    def get_rels(self, obj, rel):
        return self.get_rels_out(obj,rel) + self.get_rels_inc(obj,rel)

    def get_rel_out(self, obj, rel):
        rels=self.get_rels_out(obj,rel)
        return rels[0] if rels else None
    def get_rel_inc(self, obj, rel):
        rels=self.get_rels_inc(obj, rel)
        return rels[0] if rels else None
    def get_rel(self, obj, rel):
        rels = self.get_rels(obj,rel)
        return rels[0] if rels else None


from .imports import *

DEFAULT_SCOPE=['read']
NUM_GOT_CONTEXTS=0
NUM_GOT_STATUS=0

@lru_cache()
def Tron():
    # print('creating Tron object')
    return Mastotron()

CACHES = {}
APIS = {}

class Mastotron():
    def __init__(self):
        global CACHES, API_SERVERS

        self._api = APIS
        self._posts = {}
        self._seen_urls = set()
        self._caches = CACHES
        self._logmsg = print

    def logmsg(self,*x,**y):
        return self._logmsg(*x,**y)
    
    def _get_path_api_auth(self, server):
        return os.path.join(
            path_data, 
            'mastodon_clients',
            get_server_name(server)+'.secret'
        )
    def _get_path_user(self, account_name):
        un,server = parse_account_name(account_name)
        return os.path.join(
            path_data,
            f'{un}@{server}'
        )
        
    def _get_path_user_auth(self, account_name):
        return os.path.join(
            self._get_path_user(account_name),
            f'usercred.secret'
        )

    def api(self, server_or_account):
        return self.api_server(server_or_account) if not '@' in server_or_account else self.api_user(server_or_account)

    def api_server(self, server):
        server = get_server_name(server)
        path_client_secret = self._get_path_api_auth(server)

        if not server in self._api:
            odir=os.path.dirname(path_client_secret)
            if not os.path.exists(odir): os.makedirs(odir)
            
            Mastodon.create_app(
                f'mastotron_{server.replace(".","")}',
                api_base_url = f'https://{server}/',
                to_file = path_client_secret
            )

            self._api[server] = Mastodon(client_id = path_client_secret)
        return self._api[server]

    
    def user_is_init(self, account_name):
        path_user_secret = self._get_path_user_auth(account_name)
        return os.path.exists(path_user_secret)

    def user_auth_url(self, account_name, scopes=DEFAULT_SCOPE):
        un,server = parse_account_name(account_name)
        api = self.api(server)
        return api.auth_request_url(scopes=scopes)

    def api_user(self, account_name, code=None, direct_input=False, scopes=DEFAULT_SCOPE):
        un,server = parse_account_name(account_name)
        path_user_secret = self._get_path_user_auth(account_name)
        if not os.path.exists(path_user_secret):
            odir = os.path.dirname(path_user_secret)
            if not os.path.exists(odir): os.makedirs(odir)
            
            api = self.api(server)
        
            if not code:
                url=api.auth_request_url(scopes=scopes)
                if not direct_input:
                    return url
                else:
                    webbrowser.open(url)
                    code = input('Paste the code from the browser:\n')
            
            api.log_in(code = code, to_file = path_user_secret, scopes=DEFAULT_SCOPE)
            return api
        else:
            return Mastodon(access_token = path_user_secret)
    
    @cached_property
    def db(self): 
        return TinyDB(path_tinydb)
    
    def cache(self, dbname='cachedb'):
        return SqliteDict(path_db+'.'+dbname, autocommit=True)
    
    def status(self, url_or_uri, **post_d):
        global NUM_GOT_STATUS

        res = {}
        url = url_or_uri
        if url:
            with self.cache('status') as cache:
                if post_d:
                    cache[url]={**cache.get(url,{}), **post_d}
                    res = cache[url]
                elif url in cache:
                    res = cache[url]
                else:
                    NUM_GOT_STATUS+=1
                    # if NUM_GOT_STATUS>1: STOPXX

                    try:
                        server_name = get_server_name(url)
                        status_id = get_status_id(url)
                        api = self.api_server(server_name)

                        # print(f'getting {server_name}\'s {status_id} from API')
                        res = api.status(status_id)
                        log.debug(f'got {url} from API')
                        cache[url] = res
                    except Exception as e:
                        # log.error(f'!! {e}: {url} !!')
                        pass
        return res
        
    def post(self, url_or_uri, **post_d):
        if not url_or_uri: return
        if not url_or_uri in self._posts:
            self._posts[url_or_uri] = PostModel({
                **self.status(url_or_uri, **post_d),
                **{'_id':url_or_uri}
            })
        return self._posts[url_or_uri]
            

    def status_context(self, uri, **status_d):
        global NUM_GOT_CONTEXTS
        # with self.cache('context') as cache:
        cache = self.cache('context')
        if status_d:
            # cache[uri] = {**cache.get(uri,{}), **status_d}
            cache[uri] = status_d
        else:
            if not uri in cache:
                server,uname,status_id = get_server_account_status_id(uri)
                # print(server,uname,status_id)
                NUM_GOT_CONTEXTS+=1
                # if NUM_GOT_CONTEXTS>1: STOPXXX
                try:
                    # print(f'getting {server}\'s CONTEXT {status_id} from API')
                    statd=self.api_server(server).status_context(status_id)
                    cache[uri] = statd

                    post = Post(uri)
                    l_pre=[Post(**d) for d in statd.get('ancestors',[])]
                    l_post=[Post(**d) for d in statd.get('descendants',[])]
                    # print('post:',post)
                    # pprint(l_pre)
                    # pprint(l_post)
                    
                    if l_pre:
                        l_preself=l_pre+[post]
                        for i in range(len(l_preself)-1):
                            a=l_preself[i]
                            b=l_preself[i+1]
                            if b.is_reply and not b.out(REL_IS_REPLY_TO):
                                # print(f'{a} --> {b} ???')
                                b.relate(a,REL_IS_REPLY_TO)
                    
                    if l_post:
                        a=post
                        for b in l_post:
                            if b.is_reply and not b.out(REL_IS_REPLY_TO):
                                # print(f'{a} --> {b} ???')
                                b.relate(a,REL_IS_REPLY_TO)
                    
                except Exception as e:
                    # log.error(e)
                    # print(f'!! {e} !!')
                    return {}
        return cache[uri]

    
    # @lru_cache
    def timeline_minute(
            self,
            account_name='',
            timestamp=None,
            dtobj=None,
            force=False,
            minute_blur=BLUR_MINUTES,
            save = True,
            **timeline_opts):
        
        dkey,min_id,max_id = dtimekey(
            dtobj=dtobj,
            timestamp=timestamp,
            minute_blur=minute_blur
        )

        with self.cache('timeline_minute') as cache:
            if force or not dkey in cache:
                self.logmsg(f'requesting updates for period {dkey}')
                api = self.api_user(account_name)
                timeline = api.timeline(max_id=max_id, min_id=min_id, **timeline_opts)
                # print('<<< got res',timeline)
                posts=PostList(timeline)
                if posts: 
                    self.logmsg(f'{len(posts)} posts found in {dkey}')
                if save: cache[dkey] = [p._id for p in posts]
            else:
                posts = PostList(cache[dkey])

            return posts


    def timeline_iter(
            self, 
            account_name='', 
            timestamp=None, 
            minute_blur=BLUR_MINUTES, 
            incl_now=False, 
            only_latest=False,
            lim=LIM_TIMELINE, 
            unread_only=False, 
            seen=set(), 
            max_mins=60,
            filter_func=lambda post: True,
            **kwargs):

        iterr=iter_datetimes(
            timestamp, 
            minute_blur=minute_blur, 
            max_mins=max_mins
        )
        if not incl_now: next(iterr)
        i=0
        for dtobj in iterr:
            if i>=lim: break
            posts = self.timeline_minute(
                account_name=account_name,
                dtobj=dtobj,
                minute_blur=minute_blur,
                **kwargs
            )
            if posts:
                for post in posts:
                    if i>=lim: break
                    if not unread_only or not post.is_read:
                        if not {p._id for p in post.allcopies} & seen:
                            if filter_func(post):
                                yield post
                                i+=1
                if only_latest: break
                
                
    
    def timeline_iter_mp(
            self, 
            account_name='', 
            timestamp=None, 
            minute_blur=BLUR_MINUTES, 
            incl_now=False, 
            lim=LIM_TIMELINE, 
            unread_only=False, 
            seen=set(), 
            max_mins=60,
            num_proc=4,
            filter_func=lambda post: True,
            **kwargs):

        iterr=iter_datetimes(timestamp, minute_blur=minute_blur, max_mins=max_mins)
        if not incl_now: next(iterr)

        objs = [(account_name,dtobj,minute_blur,kwargs) for dtobj in iterr]

        def _init(state):
            state['tron'] = Mastotron()

        def _task(state, account_name, dtobj, minute_blur, kwargs):
            res = state['tron'].timeline_minute(
                account_name=account_name,
                dtobj=dtobj,
                minute_blur=minute_blur,
                **kwargs
            )
            return [p._id for p in res]

        i=0
        
        from mpire.pool import WorkerPool
        with WorkerPool(n_jobs=num_proc, use_worker_state=True, start_method='threading') as pool:
            for idlist in pool.imap_unordered(_task, objs, worker_init=_init):
                for id in idlist:
                        post = Post(id)
                        if post._id not in seen and (not unread_only or not post.is_read) and filter_func(post):
                            yield post
                            i+=1
                            print(f'> [{i}] {post} ({post.datetime_str_h})')
                            if i>=lim: break
                        if i>=lim: break

    def timeline(self, account_name='', **kwargs): 
        return PostList(self.timeline_iter(account_name=account_name, **kwargs))



    def latest(self, account_name='', **kwargs):
        return PostList(
            self.timeline_iter(
                account_name=account_name,
                only_latest=True,
                **kwargs
            )
        )


from .imports import *


class Poster(DictModel):

    @property
    def account(self):
        un, server = parse_account_name(self.url)
        return f'{un}@{server}'

    @property
    def is_valid(self):
        return bool(self.uri)

    @property
    def data(self):
        return dict(
            uri=self.uri,
            account=self.account,
            name=self.display_name,
            url_local=self.url_local,
            text=self.text,
            html=self.html,
            num_followers=self.num_followers,
            num_following=self.num_following,
            is_bot=self.bot,
            is_org=self.group,
            timestamp=self.timestamp,
        )

    @property
    def timestamp(self):
        return int(round(self.created_at.timestamp()))
    
    def __hash__(self):
        return hash(self.acct)
    
    def __eq__(self, other):
        return self._id == other._id

    def __repr__(self):
        return f'Poster({self._id})'

    @cached_property
    def html(self): return self._repr_html_(allow_embedded=True)

    @cached_property
    def text(self): return unhtml(self.note).strip() if self.note else ''

    def _repr_html_(self, allow_embedded=False, **kwargs):
        return f'<span class="author"><img src="{self.avatar}" width="50" height="50" /> <a href="{self._id}" target="_blank">{self.display_name}</a> ({self.followers_count:,} 👥){self.note if allow_embedded else ""}</span>'

    @property
    def num_followers(self):
        return self.followers_count if self.followers_count else 0

    @property
    def num_following(self):
        return self.following_count if self.following_count else 0

    def get_url_local(self, server):
        return '/'.join([server,'@'+self.acct])

    @cached_property
    def node_data(self):
        odx={}
        odx['html'] = self._repr_html_(allow_embedded=True)
        odx['shape']='circularImage'
        odx['image'] = self.avatar
        odx['size'] = 25
        odx['text'] = self.display_name
        odx['node_type']='user'
        odx['color']='#111111'
        return odx


from .imports import *

class GraphDB:
    def __init__(self, account_name_or_mastotron_obj, name='graphdb'):
        self.name = name
        from .mastotron import Mastotron
        
        if type(account_name_or_mastotron_obj) is Mastotron:
            self._tron = account_name_or_mastotron_obj
        elif type(account_name_or_mastotron_obj) is str:
            self._tron = Mastotron(account_name_or_mastotron_obj)
        else:
            raise Exception('Must provide either a string for an account name or a mastotron object.')

        from cog.torque import Graph
        self._g = Graph(self.name, cog_home=os.path.basename(self.path_g), cog_path_prefix=os.path.dirname(self.path_g))
        

    @property
    def path_g(self):
        return os.path.join(self._tron.path_acct, 'cog.graph')
    @property
    def path_db(self):
        return os.path.join(self._tron.path_acct, 'tinydb.json')

    @property
    def g(self): return self._g
    @property
    def db(self): return self._db
    # d = db

    def update(self, timeline_type='home', max_posts=100, hours_ago=24):
        for post in self._tron.iter_timeline(
                timeline_type=timeline_type,
                max_posts=max_posts,
                hours_ago=hours_ago):
            self.ingest_post(post, timeline_type=timeline_type)

    def ingest_post(self, post, force=False, **extra):
        # if force or not post.uri in self.d:
        print(f'>> ingesting: {post.uri}')
        Q = Query()
        self.db.upsert(
            {**post.data, **extra},
            Q.uri == post.uri
        )
        # self.d[post.author.uri] = post.author.data
        self.db.upsert(
            post.author.data,
            Q.uri == post.author.uri
        )
        self.g.put(post.author.uri,'posted',post.uri)


from .imports import *


svg_str_htmlkey='[[HTML]]'
# svg_str=f'''<svg xmlns="http://www.w3.org/2000/svg" width="400" height="200"><rect x="0" y="0" width="100%" height="100%" fill="#7890A7" stroke-width="5" stroke="#ffffff" ></rect><foreignObject x="1" y="1" width="100%" height="100%"><div xmlns="http://www.w3.org/1999/xhtml" style="font-size:40px">{svg_str_htmlkey}</div></foreignObject></svg>'''
svg_str=f'''<svg xmlns="http://www.w3.org/2000/svg" width="800" height="400"><foreignObject x="1" y="1" width="100%" height="100%"><div xmlns="http://www.w3.org/1999/xhtml" style="font-size:4em">{svg_str_htmlkey}</div></foreignObject></svg>'''


def get_svg_url(svg_str):
    return "data:image/svg+xml;charset=utf-8," + encodeURIComponent(svg_str)

def html2svg(html):
    return get_svg_url(svg_str.replace(svg_str_htmlkey, html))




def to_html(x, **y):
    from .post import PostModel
    if isinstance(x, PostModel):
        return post_to_html(x, **y)
    else:
        return ''

#                 <p>Post ID: <a href="{self._id}" target="_blank">{self.id}</a></p>


def post_to_html(self, allow_embedded = True, url=None, local_server='', **y): 
    datestr=f'<a href="{self._id}" target="_blank">{self.datetime_str_h}</a>'
    austr=f'<img class="post_avatarimg" src="{self.author.avatar}" width="50px" height="50px" /><a href="{self._id}" target="_blank">{self.author.display_name}</a> ({self.author.followers_count:,} 👥)'
    imgs_urls = [d.get('preview_url') for d in self.media_attachments] if self.media_attachments else []
    imgs_str = "    ".join([
        f'<a href="{xurl}" target="_blank"><img class="post_postimg" src="{xurl}" /></a>'
        for xurl in imgs_urls
        if xurl
    ])
    imgs_str=f'<center>{imgs_str}</center>' if imgs_str else ''

    # class string
    classes = ['post']
    if self.is_boost: classes.append('reblog')
    if self.is_reply: classes.append('reply')
    class_str = ' '.join(classes)

    # embed
    embed_str = ''
    if allow_embedded:
        if self.is_boost and self.is_boost_of:
            embed_str = post_to_html(self.is_boost_of)
        elif self.is_reply and self.is_reply_to:
            embed_str = post_to_html(self.is_reply_to)
    
    # header
    hdrstr=f'<div class="post_author">{austr}</div>'
    uristr=f''
    # stats
    url = self.url if self.url else self.uri
    localurl = f'https://{local_server}/authorize_interaction?uri={url}' if local_server else url
    stats_str = f'''
        <div class="post_stats">
            {self.replies_count:,} 🗣
            |
            {self.reblogs_count:,} 🔁
            |
            {self.favourites_count:,} 💙
            |
            <span class="post_uristr"><a href="{self.urli}" target="_blank">URI</a></span>
            |
            <span class="post_uristr">Posted <a href="{localurl}" target="_blank">{self.datetime_str_h}</a></span>
        </div>
    ''' if not self.is_boost else f'''
        <div class="post_stats">
            <span class="post_uristr">Reposted <a href="{url}" target="_blank">{self.datetime_str_h}</a></span>
        </div>
    '''

    ohtml = f'''
        <div class="{class_str}">
            {hdrstr}
            {self.content}
            {imgs_str}
            {embed_str}
            {stats_str}
        </div>
    '''
    ohtml = '\n'.join([ln.lstrip() for ln in ohtml.split('\n')]).strip()
    return ohtml



def post_to_svg(self): 
    svg_str_pre='<svg xmlns="http://www.w3.org/2000/svg" width="800" height="400"><rect x="0" y="0" width="100%" height="100%" fill="#dddddd"/><foreignObject x="0" y="0" width="100%" height="100%"><div xmlns="http://www.w3.org/1999/xhtml" style="font-size:100px">'
    svg_str_post='</div></foreignObject></svg>'

    datestr=f'<a href="{self._id}" target="_blank">{self.datetime_str_h}</a>'
    austr=f'{self.author.display_name} ({self.author.followers_count:,} 👥)'

    ohtml = f'''
        <div class="postsvg">
            {austr}
            {datestr}
            {self.content}
        </div>
    '''
    osvg_str = svg_str_pre + ohtml + svg_str_post
    return get_svg_url(osvg_str)


from .imports import *

class PostNet:
    def __init__(self, posts):
        from .postlist import PostList
        self.posts = (
            PostList(posts)
            if type(posts)!=PostList
            else posts
        )
    
    def graph(self, local_server=''):
        g=nx.DiGraph()
        
        def ensure_node(node):
            ndata = node.get_node_data(local_server=local_server)
            nid = ndata.get('id')
            if not g.has_node(nid):
                g.add_node(nid, **ndata)
            return nid
        
        def ensure_edge(n1,n2,rel='',**eopts):
            n1id=ensure_node(n1)
            n2id=ensure_node(n2)
            if not g.has_edge(n1id, n2id):
                eopts['id']=f'{n1id}__{rel}__{n2id}'
                eopts['rel']=rel
                eopts['from']=n1id
                eopts['to']=n2id
                g.add_edge(n1id, n2id, **eopts)
            return (n1id,n2id)

        def add_post(post):
            if post.is_boost_of:
                return add_post(post.is_boost_of)
            
            ensure_node(post)
            if post.is_reply_to: 
                ensure_edge(post, post.is_reply_to, rel='replied_to')
            
        
        for post in self.posts: add_post(post)
        
        return g

    @cached_property
    def g(self): return self.graph()
        
    def to_adjmat(self, g=None):
        if not g: g = self.graph()
        return pd.DataFrame(
            nx.adjacency_matrix(g).todense(),
            columns = g.nodes,
            index = g.nodes
        )



        


#!/usr/bin/env python3

from .imports import *
from .gui.app import *


@click.command()
@click.option('-w','--web', is_flag=True, show_default=True, default=False,help='Run as local web server')
@click.option('-c','--code', is_flag=True, show_default=True, default=False,help='Run in ipython terminal')
def cli(web=False,code=False):
    if code:
        imp='from mastotron import *'
        cmd = ' && '.join([
            f'cd {path_code}',
            f'''ipython -c "{imp}; print('>>> {imp}')" -i'''
        ])
        return os.system(cmd)

    mainview()
    

if __name__=='__main__': cli()


from .imports import *


def rmfile(fn):
    try:
        os.remove(fn)
    except OSError:
        pass

def getlocurl(url,server):
    return f'https://{server}/authorize_interaction?uri={url}'

def get_api():
    global API
    return API

def set_api(mastotron_obj):
    global API
    API = mastotron_obj


def clean_account_name(acct):
    un,server=parse_account_name(acct)
    if un and server:
        return f'{un}@{server}'
    return ''

def parse_account_name(acct):
    un, server = '', ''
    acct = acct.strip()
    
    if acct and '@' in acct:
        if acct.startswith('@'):
            return parse_account_name(acct[1:])

        elif acct.count('@')>1:
            return parse_account_name(acct.split('/@',1)[-1])

        elif acct.startswith('http'):
            server, un = acct.split('/@')
            un = un.split('/')[0]
            server = server.split('://',1)[-1].split('/')[0]

        elif acct.count('@')==1:
            un, server = acct.split('@',1)
            server = server.split('/')[0]

    return un,server

def get_server_name(server):
    return server.split('://',1)[-1].split('/',1)[0]

def get_status_id(url):
    for x in reversed(url.split('/')):
        if x and x.isdigit():
            return int(x)

def get_account_name(x): return parse_account_name(x)[0]

def get_server_account_status_id(x):
    return (
        get_server_name(x),
        get_account_name(x),
        get_status_id(x)
    )

TRON=None
def get_tron():
    global TRON
    if TRON is None: 
        from .mastotron import Mastotron
        TRON=Mastotron()
    return TRON

class DictModel:
    def __init__(self, data_d={}, **kwargs):
        self._data = {**(data_d if data_d else {}), **(kwargs if kwargs else {})}

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            return self._data.get(name)

def to_uri(url): 
    if not url: return ''
    if url.count('@')>=2:
        return url
    else:
        url = url.replace('/users/','/@').replace('/statuses/','/')
        un,server = parse_account_name(url)
        status_ids = [y for y in url.split('/') if y.isdigit()]
        if status_ids:
            status_id = status_ids[0]
            return f'https://{server}/@{un}/{status_id}'
        return ''

def encodeURIComponent(x):
    from urllib.parse import quote
    return quote(x, safe="!~*'()")


def test_encodeURIComponent():
    inp='''<svg xmlns="http://www.w3.org/2000/svg" width="390" height="65"><rect x="0" y="0" width="100%" height="100%" fill="#7890A7" stroke-width="20" stroke="#ffffff" ></rect><foreignObject x="15" y="10" width="100%" height="100%"><div xmlns="http://www.w3.org/1999/xhtml" style="font-size:40px"> <em>I</em> am<span style="color:white; text-shadow:0 0 20px #000000;"> HTML in SVG!</span></div></foreignObject></svg>'''
    out='''%3Csvg%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%20width%3D%22390%22%20height%3D%2265%22%3E%3Crect%20x%3D%220%22%20y%3D%220%22%20width%3D%22100%25%22%20height%3D%22100%25%22%20fill%3D%22%237890A7%22%20stroke-width%3D%2220%22%20stroke%3D%22%23ffffff%22%20%3E%3C%2Frect%3E%3CforeignObject%20x%3D%2215%22%20y%3D%2210%22%20width%3D%22100%25%22%20height%3D%22100%25%22%3E%3Cdiv%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F1999%2Fxhtml%22%20style%3D%22font-size%3A40px%22%3E%20%3Cem%3EI%3C%2Fem%3E%20am%3Cspan%20style%3D%22color%3Awhite%3B%20text-shadow%3A0%200%2020px%20%23000000%3B%22%3E%20HTML%20in%20SVG!%3C%2Fspan%3E%3C%2Fdiv%3E%3C%2FforeignObject%3E%3C%2Fsvg%3E'''
    assert encodeURIComponent(inp) == out






def find_local_url(*urls):
    urls = [x for x in urls if type(x)==str and x]
    urls.sort(key=lambda x: -x.count('@'))
    return urls[0] if urls else ''

def find_localremote_url(url1,url2):
    urls = [url1,url2]
    urls.sort(key=lambda x: -x.count('@'))
    return tuple(urls)




def get_datetime_str():
    return str(dt.datetime.now()).split('.',1)[0]
def get_time_str():
    return str(dt.datetime.now()).split('.',1)[0].split()[-1]

def get_graphtime_str(timestamp=None, minute_blur=BLUR_MINUTES):
    now=get_now(timestamp)
    minute=now.minute // minute_blur * minute_blur
    upminute=minute+minute_blur
    # return f'{now.year:04}-{now.month:02}-{now.day:02} {now.hour:02}:{minute:02}'

    return f'{now.year:04}/{now.month:02}/{now.day:02} {now.hour:02}:{minute:02}-{upminute:02}'
    
def get_now(timestamp=None):
    return (
        dt.datetime.fromtimestamp(timestamp) 
        if timestamp
        else dt.datetime.now()
    )

def iter_graphtimes(timestamp=None, minute_blur=BLUR_MINUTES, max_days=7):
    now = get_now(timestamp)
    current_time = now
    while True:
        yield get_graphtime_str(timestamp=current_time.timestamp())
        current_time -= dt.timedelta(minutes=minute_blur)
        if abs(now - current_time) > dt.timedelta(days=max_days):
            break


def iter_datetimes(timestamp=None, minute_blur=BLUR_MINUTES, max_mins=7):
    now = get_now(timestamp)
    current_time = blurtime(now)
    while True:
        yield current_time
        current_time -= dt.timedelta(minutes=minute_blur)
        if abs(now - current_time) > dt.timedelta(minutes=max_mins):
            break

def get_datetimes(**kwargs): return list(iter_datetimes(**kwargs))

def blurtime(dtobj, minute_blur=BLUR_MINUTES):
    return dt.datetime(
        year=dtobj.year,
        month=dtobj.month,
        day=dtobj.day,
        hour=dtobj.hour,
        minute=dtobj.minute // minute_blur * minute_blur,
    )


def find_urls(string):
    import re

    # findall() has been used
    # with valid conditions for urls in string
    regex = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
    url = re.findall(regex, string)
    return [x[0] for x in url]



# def query_user_min_max(acct, ):
#     Tron().api_user(acct)

def dtimekey(dtobj=None,timestamp=None,minute_blur=BLUR_MINUTES):
    if dtobj or timestamp:
        if not dtobj: dtobj=dt.datetime.fromtimestamp(timestamp)

        dtime1 = blurtime(dtobj, minute_blur)
        dtime2 = dtime1 + dt.timedelta(minutes=minute_blur)        
    else:
        dtime2 = blurtime(dt.datetime.now(), minute_blur)
        dtime1 = dtime2 - dt.timedelta(minutes=minute_blur)

    dkey = get_graphtime_str(timestamp = dtime1.timestamp(), minute_blur=minute_blur)
    min_id = ( int( dtime1.timestamp() ) << 16 ) * 1000
    max_id = ( int( dtime2.timestamp() ) << 16 ) * 1000
    return (dkey,min_id,max_id)

        


## internal imports
import os,sys
path_app = os.path.realpath(__file__)
path_code = os.path.abspath(os.path.join(path_app,'..','..'))
path_codedir = os.path.abspath(os.path.join(path_code,'..'))
sys.path.insert(0,path_codedir)
from mastotron import *
from mastotron import __version__ as vnum


## external imports
import socket,time,webbrowser,signal
import gevent
import webview
import threading
from engineio.payload import Payload
Payload.max_decode_packets = 500
from threading import Lock,Thread,Event
from flask import Flask, render_template, request, redirect, url_for, session, current_app
from flask_session import Session
from flask_socketio import SocketIO, send, emit
from mastodon import StreamListener
import os,time,sys,random
from engineio.async_drivers import gevent as gevent_async_driver
import logging
log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)


## constants
MAX_UPDATE=15
CRAWL_LIM=10
CRAWL_WAIT_SEC=60
SEEN=set()
CRAWL_STARTED={}
LSTNR_STARTED={}
new_urls = set()
new_msgs = []
ACCT = None


# @lru_cache
def Tron():
    obj = Mastotron()
    obj._logmsg = logmsg_tron
    return obj

def logmsg_tron(*x,**y):
    global new_msgs
    new_msgs.append(' '.join(str(xx) for xx in x))
    # print('>>>',new_msgs[-1])

def logmsg(*x,**y):
    emitt('logmsg',' '.join(str(xx) for xx in x))
    gevent.sleep(0.1)
def logsuccess(x): emitt('logsuccess',str(x))
def logerror(x): emitt('logerror',str(x))

def emitt(key,val,*vals,broadcast=True,**opts):
    # emit(key,val,*vals,broadcast=broadcast,**opts)
    socketio.emit(key,val,*vals,broadcast=broadcast,**opts)


#####
HOSTPORTURL=f'http://{HOST}:{PORT}'

LINE1=f'MASTOTRON v{vnum}'
LINE2=f'{HOSTPORTURL}'
LINE3='(Visit this URL in your browser.'
LINE4='It\'s been copied to your clipboard.)'
# LINE4='(To exit, hold Ctrl+C or close terminal window.)'
LOGO=r'''                      _        _                   
                     | |      | |                  
  _ __ ___   __ _ ___| |_ ___ | |_ _ __ ___  _ __  
 | '_ ` _ \ / _` / __| __/ _ \| __| '__/ _ \| '_ \ 
 | | | | | | (_| \__ \ || (_) | |_| | | (_) | | | |
 |_| |_| |_|\__,_|___/\__\___/ \__|_|  \___/|_| |_|
'''

ELE=r'''
                 /eeeeeeeeeee\ 
   /RRRRRRRRRR\ /eeeeeeeeeeeee\ /RRRRRRRRRR\ 
  /RRRRRRRRRRRR\|eeeeeeeeeeeee|/RRRRRRRRRRRR\ 
 /RRRRRRRRRRRRRR +++++++++++++ RRRRRRRRRRRRRR\ 
|RRRRRRRRRRRRRR ############### RRRRRRRRRRRRRR| 
|RRRRRRRRRRRRR ######### ####### RRRRRRRRRRRRR| 
 \RRRRRRRRRRR ######### ######### RRRRRRRRRR/ 
   |RRRRRRRRR ########## ######## RRRRRRRR| 
  |RRRRRRRRRR ################### RRRRRRRRR| 
               ######     ###### 
               #####       ##### 
               #nnn#       #nnn#
'''


lwid=max(len(lstr) for lstr in ELE.split('\n'))
WELCOME_MSG = f'''
{ELE.center(lwid)}
{LINE1.center(lwid)}

{LINE2.center(lwid)}

{LINE3.center(lwid)}
{LINE4.center(lwid)}
'''


app = Flask(__name__, static_folder=path_static, template_folder=path_templates)
app.config['SESSION_PERMANENT'] = True
app.config['SESSION_TYPE'] = 'filesystem'
app.config['SESSION_FILE_DIR'] = os.path.join(path_data, 'flask_session')
Session(app)
# socketio = SocketIO(app, manage_session=False, async_handlers=False)
socketio = SocketIO(app, manage_session=False, async_mode="gevent")
# socketio = SocketIO(app, manage_session=False, async_mode="eventlet")


@app.route("/")
@app.route("/hello/")
@app.route("/hello-vue/")
def postnet(): 
    global ACCT
    acct=session.get('acct')
    if not acct or not Tron().user_is_init(acct): 
        return render_template('login.html', acct=acct, config_d=get_config())
    else:
        ACCT = acct
        return render_template('postnet.html', acct=acct, config_d=get_config())

@app.route('/hello//flaskwebgui-keep-server-alive')
@app.route('/hello/flaskwebgui-keep-server-alive')
@app.route('/flaskwebgui-keep-server-alive')
def keepalive():
    return {'status':200}


def get_config(d={}):
    defaultd={
        'LIM_NODES_GRAPH':12,
        'LIM_NODES_STACK':100,
        'DARKMODE':1,
        'VNUM':vnum,
    }
    sessiond={k:v for k,v in session.items() if k and k[0]!='_'}
    outd = {**defaultd, **sessiond, **d}
    return outd

@socketio.event
def req_config(d={}):
    emitt('res_config',get_config(d))

@socketio.event
def set_config(k,v):
    # print(f'setting config: {k} -> {v}')
    # print(f'{k} in config now =',get_config().get(k))
    session[k]=v

@socketio.event
def set_acct_name(data):
    tron = Tron()
    acct = clean_account_name(data.get('acct'))
    if acct:
        session['acct'] = acct
        if not tron.user_is_init(acct):
            try:
                url = tron.user_auth_url(acct)
                logmsg('Please enter activation code')
                emitt('get_auth_url', {'acct':acct,'url':url} )
            except Exception as e:
                un,server=parse_account_name(acct)
                emitt('server_not_giving_code', {'server':server, 'acct':acct})
    else:
        emitt('invalid_user_name', data)
        

@socketio.event
def do_login(data):
    tron = Tron()
    acct = data.get('acct')
    code = data.get('code')
    if acct and code:
        tron.api_user(acct, code=code, direct_input=False)
        if tron.user_is_init(acct):
            logmsg(f'{acct} logged in')
            emitt('login_succeeded', dict(acct=acct))
        else:
            logmsg(f'{acct} NOT logged in')
            emitt('login_failed', dict(acct=acct))


def get_acct_name(data={}): return data.get('acct', ACCT)
def get_srvr_name(data={}): return parse_account_name(get_acct_name(data))[-1]



#########
# GRAPH #
#########

class NodeListener(StreamListener):
    def __init__(self,acct,*x,**y):
        super().__init__(*x,**y)
        self.acct=acct

    def on_update(self, status):
        url = status.get('url')
        post = Post(url, **status)
        if post:
            update_posts([post], acct=self.acct, force_push=True)

@socketio.event
def start_updates(data={}):
    # print('!!! starting updates !!!')
    acct = get_acct_name(data)
    start_listener(acct)
    gevent.sleep(30)
    start_crawler(acct)
    
def start_crawler(acct):
    if not acct: return
    if acct in CRAWL_STARTED: CRAWL_STARTED[acct].stop()
    CRAWL_STARTED[acct] = crawler = Crawler(acct)
    socketio.start_background_task(crawler.run)

def start_listener(acct):
    if not acct: return
    if acct in LSTNR_STARTED: LSTNR_STARTED[acct].close()

    LSTNR_STARTED[acct]=Tron().api_user(acct).stream_user(
        NodeListener(acct),
        run_async=True
    )



class Crawler():
    def __init__(self, acct, on=True):
        self.acct=acct
        self.on = on
    
    def crawl(self, lim=CRAWL_LIM, sec_between=.1):
        global SEEN
        if not self.on: return
        iterr=Tron().timeline_iter(self.acct, seen=SEEN, max_mins=60*24, lim=lim)
        for i,post in enumerate(iterr):
            if not self.on: return
            print(i,post)
            update_posts([post], acct=self.acct, bg=True)
            gevent.sleep(sec_between)
    
    def wait(self, sec=CRAWL_WAIT_SEC):
        if not self.on: return
        gevent.sleep(sec)
    
    def run(self):
        while self.on:
            self.crawl()
            self.wait()

    def stop(self):
        self.on = False
        gevent.sleep(0.001)



@socketio.event
def get_updates(data={}):
    global SEEN
    acct = get_acct_name(data)
    if not acct: return
    SEEN|=set(data.get('ids_now',[]))
    lim = data.get('lim') if data.get('lim') else get_config().get('LIM_NODES_STACK')
    force_push = data.get('force_push',False)
    only_latest = data.get('only_latest',False)
    max_mins = data.get('max_mins',60*24)
    unread_only=data.get('unread_only',True)
    bg=data.get('bg',False)

    lim=lim if lim<MAX_UPDATE else MAX_UPDATE
    iterr=Tron().timeline_iter(
    # tl=Tron().timeline(
        acct, 
        unread_only=unread_only,
        seen=SEEN,
        max_mins=max_mins,
        lim=lim,
        only_latest=only_latest
    )
    # update_posts(tl, ids_done=SEEN, force_push=force_push)
    # SEEN|={p._id for post in tl for p in post.allcopies}
    posts=[]
    for i,post in enumerate(iterr):
        if len(posts) < lim:
            if lim>1:
                logmsg(f'found {i+1} of {lim}: {post} ({post.datetime_str_h})')
            else:
                logmsg(f'found: {post} ({post.datetime_str_h})')
            posts.append(post)

            update_posts([post], ids_done=SEEN, force_push=force_push, bg=bg, acct=acct)
        else:
            break
    
    # update_posts(posts, ids_done=SEEN, force_push=force_push, bg=bg)
    SEEN|={p._id for post in posts for p in post.allcopies}

        # update_posts(
        #     [post],
        #     ids_done=SEEN, 
        #     force_push=force_push
        # )
        # SEEN|={p._id for p in post.allcopies}



def update_posts(tl, omsg='refreshed', emit_key='get_updates',ids_done=None,unread_only=True, force_push=False, bg=False, acct=None, force_push_once=False):
    global SEEN
    if type(tl)!=PostList: tl=PostList(tl)
    SEEN|={p._id for p in tl}
    if len(tl):
        # print('>',tl)
        tnet = tl.network()
        local_server = parse_account_name(acct)[-1] if acct else get_srvr_name()
        nx_g = tnet.graph(local_server=local_server)
        nx_g.remove_nodes_from(
            n 
            for n,d in list(nx_g.nodes(data=True))
            if (unread_only and d.get('is_read'))
            or (ids_done and d.get('id') in ids_done)
        )
        nodes = [d for n,d in nx_g.nodes(data=True)]
        edges = [d for a,b,d in nx_g.edges(data=True)]

        # print('ne',[d['id'] for d in nodes],[d['id'] for d in edges])

        omsg = f'{len(nodes)+len(edges)} new updates @ {get_time_str()}'
        if nodes or edges:
            odata = dict(nodes=nodes, edges=edges, logmsg=omsg, force_push=force_push, bg=bg, force_push_once=force_push_once)
            # for n in nodes: print('++',n['id'])
            emitt(emit_key, odata)
            # time.sleep(0.1)
            gevent.sleep(.1)
            return True
    
    # omsg = f'no new updates @ {get_time_str()}'
    # logmsg(omsg)
    return False

@socketio.event
def add_context(node_id):
    # print('add_context',node_id)
    post = Post(node_id)
    if post:# unread_convo = PostList(p for p in post.convo if not p.is_read)
        convo = post.convo
        update_posts(convo[:LIM_TIMELINE], force_push_once=True)









@socketio.event
def set_darkmode(darkmode):
    key='DARKMODE'
    session[key] = int(darkmode)

@socketio.event
def set_in_session(dict):
    for k,v in dict.items():
        session[k]=v

@socketio.event
def mark_as_read(node_ids):
    # print('mark_as_read', node_ids)
    for node_id in node_ids:
        post = Post(node_id)
        post.mark_read()









class OpenBrowser(Thread):
    def __init__(self):
        super(OpenBrowser, self).__init__()
    def notResponding(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        return sock.connect_ex((HOST, PORT))
    def run(self):
        while self.notResponding():
            # print('Did not respond')
            gevent.sleep(0.1)
        # print('Responded')
        webbrowser.open_new(f'http://{HOST}:{PORT}/') 


def send_update(nodes=[], edges=[]):
    odata=dict(nodes=nodes, edges=edges)
    emitt('get_updates', odata)
    return odata


def view(**kwargs):
    OpenBrowser().start()

def mainview(**kwargs):
    # view(**kwargs)
    # main(**kwargs)
    browse(app = app, func = main)


def main(debug=False, **kwargs): 
    try:
        import pyperclip
        pyperclip.copy(HOSTPORTURL)
    except Exception:
        pass
    print(WELCOME_MSG)
    try:
        # socketio.start_background_task(crawl_updates)
        # socketio.start_background_task(browse, app=app)
        return socketio.run(
            app, 
            debug=debug, 
            # allow_unsafe_werkzeug=True,
            port=PORT,
            host=HOST
        )
    # except (KeyboardInterrupt,EOFError):
    except AssertionError:
        print('goodbye')
        exit()



def browse(app=None,**kwargs):
    try:
        from screeninfo import get_monitors
        m = get_monitors()[0]
        width = m.width
        height = m.height
    except Exception:
        width = 1400
        height = 1100

    with app.app_context():
        gevent.sleep(1)
        webview.create_window(
            title=f'Mastotron {vnum}', 
            url=f'http://{HOST}:{PORT}',
            fullscreen=False,
            width=width,
            height=height,
            min_size=(400,300),
            frameless=False,
            easy_drag=True,
            text_select=True,
            confirm_close=True,
        )
        return webview.start(
            private_mode=False,
            storage_path=path_srvr,
            debug=True,
            **kwargs
        )



if __name__=='__main__': mainview(debug=False)


from setuptools import setup
from setuptools import find_packages
pkgs=find_packages(exclude=('tests',))
with open('README.md') as f: readme=f.read()
with open('requirements.txt') as f: reqs=[x.strip() for x in f.read().split('\n') if x.strip()]

setup(
    name='mastotron',
    version='0.4.4',
    url=f"https://github.com/quadrismegistus/mastotron",
    license='MIT',
    author="Ryan Heuser",
    author_email="ryan.heuser@princeton.edu",
    description="Algorithmifying mastodon",
    long_description=readme,
    long_description_content_type="text/markdown",

    packages=pkgs,

    install_requires=reqs,

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],


    py_modules=['mastotron'],
    entry_points={
        'console_scripts': [
            f'mastotron = mastotron.cli:cli',
        ],
    },
    package_data={
        '': ['static/*', 'templates/*'],
    }
)


from .imports import *
from .postlist import PostList
from .htmlfmt import to_html
from .db import TronDB
from .htmlfmt import html2svg, post_to_svg

is_reply_to_uri_key = 'is_reply_to__id'

def Post(url_or_uri='', **post_d):
    if type(url_or_uri) is PostModel: return url_or_uri
    if type(url_or_uri) in {AttribAccessDict,dict}:
        post_d={**url_or_uri, **post_d}
        url_or_uri=''
    
    if not url_or_uri:
        if not post_d: return
        url_or_uri = post_d.get('url') if post_d.get('url') else post_d.get('uri')
        if not url_or_uri: return
    from .mastotron import Tron
    post = Tron().post(url_or_uri, **post_d)
    return post.source if post else post


@total_ordering
class PostModel(DictModel):
    def __init__(self,*x,**y):
        super().__init__(*x,**y)
        self.boot()

    def __eq__(self, post):
        post=Post(post)
        return type(post) is PostModel and self._id==post._id

    def __hash__(self):
        return hash(self._id)

    def __gt__(self, other):
        return self.timestamp > other.timestamp
    
    def __str__(self): return self._id
    def __repr__(self): return f"Post('{self._id}')"
    def _repr_html_(self): return self.html

    ############
    ### BOOT ###
    ############

    def boot(self):
        # time
        if not self.out(REL_GRAPHTIME):
            self.store_graphtime()
        
        # num id
        if not self.out(REL_HAS_ID):
            self.relate(self.au_id, REL_HAS_ID)

        # replies
        if self.is_reply and not self.out(REL_TO_ID):
            au_id = f'{self.in_reply_to_account_id}__{self.in_reply_to_id}'
            self.relate(au_id, REL_TO_ID)

        # local/remote
        # self.is_local_for
        
    


    ############
    ### TIME ###
    ############

    @cached_property
    def graphtime(self):
        return get_graphtime_str(self.timestamp)
    @property
    def graphtime_iter(self):
        yield from iter_graphtimes(self.timestamp)
    
    @property
    def graphtime_indb(self):
        return self.out(REL_GRAPHTIME)
    def store_graphtime(self):
        if not self.graphtime_indb:
            self.relate(self.graphtime, REL_GRAPHTIME)

    #############
    ### LOCAL ###
    #############

    @cached_property
    def is_local_for(self):
        from .mastotron import Tron
        if self.is_local:
            out = Tron().post(self.out(REL_IS_LOCAL_FOR))
            if not out:
                out = Tron().post(self.urli, **self._data)
                self.relate(out,REL_IS_LOCAL_FOR)
            return out

    @cached_property
    def source(self):
        return self.is_local_for if self.is_local_for else self
    @property
    def is_source(self):
        return self.source == self
    
    @cached_property
    def au_id(self):
        return f'{self.author.id}__{self.id}'
    
    @cached_property
    def copies(self):
        return PostList(self.incs(REL_IS_LOCAL_FOR))
    @cached_property
    def local(self):
        if self.is_local: return self
        if self.copies: return self.copies[0]
    
    @cached_property
    def allcopies(self):
        l=[self.source] + list(self.source.copies) + [self.source.is_boost_of] + list(self.source.was_boosted_by)
        return {x for x in l if x is not None and type(x) is PostModel}
    
    def get_local(self, server):
        copies = [p for p in self.source.copies if p.localserver==server]
        return copies[0] if copies else None

    @property
    def localsource(self):
        if self.local: return self.local
        return self.source
    def get_localsource(self, server):
        copies = [p for p in self.source.copies if p.localserver==server]
        return copies[0] if copies else self.source
    
    
    #############
    ### BOOST ###
    #############
    
    @cached_property
    def is_boost_of(self):
        if self.is_boost:
            out = Post(self.out(REL_IS_BOOST_OF))
            if not out:
                url = self.reblog.get('url')
                out = Post(url, **dict(self.reblog))
                self.relate(out,REL_IS_BOOST_OF)
            return out

    @cached_property
    def was_boosted_by(self):        
        if not self.is_boost:
            return PostList(self.incs(REL_IS_BOOST_OF))
        return PostList()
    


    #############
    ### REPLY ###
    #############

    @cached_property
    def is_reply_to(self):
        if self.is_reply:
            out = self._get_is_reply_to_by_rel()
            if not out:
                out = self._get_is_reply_to_by_au_id()
                if not out:
                    self.get_context()
                    out = self._get_is_reply_to_by_au_id()
            return out

    def _get_is_reply_to_by_rel(self):
        return Post(self.out(REL_IS_REPLY_TO))

    def _get_is_reply_to_by_au_id(self):
        out_id = self.out(REL_TO_ID)
        if out_id: 
            res = TronDB().gdb.v(out_id).inc(REL_HAS_ID).all().get('result')
            if res:
                out_uri = res[0].get('id')
                out = Post(out_uri)
                self.relate(out, REL_IS_REPLY_TO)
                return out
    
    @cached_property
    def was_replied_to(self):
        # self.context
        return PostList(self.incs(REL_IS_REPLY_TO))

    ##########
    # READ?? #
    ##########

    def mark_read(self):
        for post in self.allcopies:
            # TronDB().unrelate(self, NODE_READ_STATUS_IS_UNREAD, REL_READ_STATUS)
            post.relate(NODE_READ_STATUS_IS_READ, REL_READ_STATUS)
    
    def mark_unread(self):
        for post in self.allcopies:
            # TronDB().unrelate(self, NODE_READ_STATUS_IS_READ, REL_READ_STATUS)
            post.relate(NODE_READ_STATUS_IS_UNREAD, REL_READ_STATUS)
    
    @property
    def is_read(self):
        for post in self.allcopies:
            nstatus=post.out(REL_READ_STATUS)
            if nstatus is not None:
                return nstatus == NODE_READ_STATUS_IS_READ
    
    



    @property
    def urli(self):
        return to_uri(self.url if self.url else self.uri)

    @property
    def is_valid(self):
        return bool(self._id) and bool(self.author) and bool(self.author._id)

    @cached_property
    def server(self):
        return parse_account_name(self.url)[-1]
    @cached_property
    def localserver(self):
        return self._id.split('://',1)[-1].split('/')[0]
    @cached_property
    def un(self):
        return parse_account_name(self.url)[0]
    @cached_property
    def acct(self):
        return f'{self.un}@{self.server}'


    def api(self):
        from .mastotron import Tron
        return Tron().api_server(self.server)
    

    @property
    def author(self):
        from .poster import Poster
        return Poster(self.account, _id=self.acct)

    def get_url_local(self, server):
        from .mastotron import Tron
        cache=Tron().cache('url_to_uri')
        return cache.get(f'{self._id}__in__{server}')
    
    @property
    def timestamp(self):
        return self._data.get(
            'timestamp',
            self.datetime.timestamp()
        )
    @property
    def timestamp_bitshift(self):
        return (self.id >> 16) / 1000

    @property
    def datetime_bitshift(self):
        return dt.datetime.fromtimestamp(self.timestamp_bitshift)

    @cached_property
    def datetime(self):
        if type(self.created_at) is str:
            from dateutil.parser import parse
            return parse(self.created_at)
        else:
            return self.created_at
    
    @property
    def datetime_str(self):
        return self.datetime.strftime("%m/%d/%Y at %H:%M:%S")
        
    @cached_property
    def post_id(self):
        return os.path.join(self.poster_id, self.url.split('/')[-1])
    
    @cached_property
    def status_id(self):
        return self.url.split('/')[-1]

    @cached_property
    def poster_id(self):
        return self.author.account
    
    # @cached_property
    # def status_context(self):
    #     from .mastotron import Tron
    #     tron = Tron()

    #     res = tron.status_context(self._id)
    #     if not res: return [], []

    #     posts_pre=res.get('ancestors',[])
    #     posts_post=res.get('descendants',[])

    #     ## quick store?
    #     for status_d in posts_pre+posts_post:
    #         url = status_d.get('url')
    #         tron.status(url, **status_d)
    #         tron.status_context(url, **res)

    #     return posts_pre, posts_post

    @cache
    def status_context_d(self, key='id'):
        pre,post=self.status_context
        return dict((d.get(key),d) for d in pre+[self._data]+post)

    @property
    def is_reply_to_status(self):
        o={}
        if self.is_reply and self.in_reply_to_id:
            o=self.status_context_d().get(self.in_reply_to_id)
            if not o:
                pre,post=self.status_context
                if pre:
                    o=pre[-1]
        return o


    def iter_context(self):
        posts_pre,posts_post = self.status_context
        def yield_dicts():
            yield from reversed(posts_pre)
            yield from posts_post
        def yield_posts():
            yield self
            for d in yield_dicts(): 
                post = Post(**d)
                yield post
        yield from yield_posts()

    def iter_contexts(self,**kwargs):
        for post in self.allcopies:
            yield from post.iter_context(**kwargs)

    # def get_context(self, lim=None, **kwargs):
    #     return PostList(self.iter_contexts(**kwargs), lim=lim)

    def get_context(self, **kwargs):
        from .mastotron import Tron
        statd=Tron().status_context(self._id)
        l_pre=statd.get('ancestors',[])
        l_post=statd.get('descendants',[])
        def topost(d): return Post(d.get('url'), **d)
        l=list(map(topost,l_pre)) + [self] + list(map(topost,l_post))
        return PostList(l,**kwargs)

    @cached_property
    def context(self): 
        # return PostList(subpost for post in self.allcopies for subpost in post.get_context())
        return self.get_context()


    def relate(self, obj, rel):
        if obj and rel:
            return TronDB().relate(self, obj, rel)
    
    def inc(self, rel): return TronDB().get_rel_inc(self,rel)
    def incs(self, rel): return TronDB().get_rels_inc(self,rel)
    def out(self, rel): return TronDB().get_rel_out(self,rel)
    def outs(self, rel): return TronDB().get_rels_out(self,rel)
    def alls(self, rel): return self.incs(rel) + self.outs(rel)
    
    def rels_exist(self):
        v=TronDB().gdb.v(self._id)
        incs=v.inc().all().get('result')
        outs=v.out().all().get('result')
        return bool(incs) or bool(outs)
    
    
    @cached_property
    def replies(self): return self.was_replied_to

    @property
    def convo(self):
        def iterr():
            yield self
            context = self.context
            ids=set(self.alls(REL_IS_REPLY_TO))
            for post in context:
                if post._id in ids or post.source._id in ids:
                    yield post
        return PostList(iterr())
    

    @cached_property
    def op(self):
        if self.is_reply_to:
            print(f'searching for OP in {self.is_reply_to}')
            return self.is_reply_to.op
        return self


    def iter_reply_chain(self):
        yield self
        if self.is_reply_to:
            yield from self.is_reply_to.iter_reply_chain()
    
    @cached_property
    def reply_chain(self):
        return PostList(self.iter_reply_chain())
    
    @property
    def unread_reply_chain(self):
        return PostList(post for post in self.reply_chain if not post.is_read)
                    
    

    def get_html(self, allow_embedded=False, local_server=''):
        from .htmlfmt import post_to_html
        return post_to_html(
            self, 
            allow_embedded=allow_embedded,
            local_server=local_server
        )
    
    @property
    def num_reblogs(self):
        if False: # self.is_boost and self.is_boost_of: 
            res = self.is_boost_of.num_reblogs
        else:
            res = self.reblogs_count
        return res if res else 0
    
    @property
    def num_likes(self):
        if False: #self.is_boost and self.is_boost_of: 
            res = self.is_boost_of.num_likes
        else:
            res = self.favourites_count
        return res if res else 0
    
    @property
    def num_replies(self):
        if False: #self.is_boost and self.is_boost_of: 
            res = self.is_boost_of.num_replies
        else:
            res = self.replies_count
        return res if res else 0

    @property
    def is_boost(self):
        return self.reblog is not None
    @property
    def is_reply(self):
        return self.in_reply_to_id is not None
    @property
    def is_local(self):
        return self.urli != self._id

    @property
    def datetime_str(self):
        return self.datetime.strftime("%m/%d/%Y at %H:%M:%S")

    @property
    def datetime_str_h(self):
        import human_readable as humr
        return humr.date_time(
            datetime.now().timestamp() - self.timestamp
        )
    

    @cached_property
    def scores(self):
        """
        These scoring ideas come from mastodon_digest (https://github.com/hodgesmr/mastodon_digest).
        """
        scores={}

        # simple score
        scores['Simple'] = gmean([
            self.num_reblogs+1, 
            self.num_likes+1,
        ])

        # extended simple score
        scores['ExtendedSimple'] = gmean([
            self.num_reblogs+1, 
            self.num_likes+1,
            self.num_replies+1
        ])

        # add weighted versions
        for score_name in list(scores.keys()):
            scores[score_name+'Weighted'] = scores[score_name] / sqrt(self.author.num_followers+1)

        scores['All'] = gmean(list(scores.values()))
        return scores

    @property
    def score(self):
        return self.get_score()

    def get_score(self, score_type='All'):
        return self.scores.get(score_type,np.nan)

    @property
    def text(self):
        import html
        return unhtml(html.unescape(self.content)).strip()

    @property
    def label(self): return self.get_label()

    def get_label(self, limsize=20, max_lines=3, **kwargs):
        text = self.text
        for url in find_urls(text): text=text.replace(url,url[:20])
        text=' '.join([w for w in text.split() if w and w[0]!='@'])
        lines=[]
        if text:
            lines = textwrap.wrap(text,limsize)
            if max_lines and len(lines)>max_lines:
                lines=lines[:max_lines]
                lines[-1]+='...'
        stext='\n'.join(lines)        
        return stext


    def get_label_orig(self, limsize=40, limwords=4, replace_urls=True):
        import html
        #stext = self.spoiler_text
        #if not stext: 
        stext=' '.join(w for w in self.text.split() if w and w[0]!='@')
        stext=html.unescape(stext).strip()
        if limwords: stext=' '.join(stext.split()[:limwords])
        if replace_urls: stext=' '.join(w.split('://',1)[-1][:20] if w.startswith('http') else w for w in stext.split())
        if limsize: stext=stext[:limsize]

        return stext.strip()

    def save(self):
        # data = {str(k):str(v) for k,v in self.data.items()}
        cache=self.cache
        from .db import TronDB
        # log.debug(f'saving: {self}')
        TronDB().set_post(self.data)
        # TronDB().gdb.updatej(data)


    
        
    
    @property
    def data_default(self):
        return {'_id':self._id, 'is_read':False}

    @cached_property
    def data(self):
        od={
            **self.data_default,
            **self._data,
            **{'timestamp':self.timestamp},
            **{f'score_{k}':v for k,v in self.scores.items()},
            **{'_id':self._id},
        }
        # if self.source: od={**od, **self.source.data}
        return od

    def get_node_data(post, local_server=''):
        odx={}
        odx['_id'] = post._id
        odx['id'] = post._id
        odx['html'] = post.get_html(allow_embedded=False, local_server=local_server)
        odx['url_local'] = getlocurl(post._id, local_server) if local_server else post._id
        odx['shape'] = 'circularImage' # if not same_author else 'box'
        # odx['shape'] = 'image'
        odx['image'] = post.author.avatar
        odx['author'] = post.author.account
        odx['author_name'] = post.author.display_name
        # odx['image'] = post_to_svg(post)
        # odx['label']=post.get_label(limsize=70, limwords=7) if not post.is_boost else 'RT'
        odx['label']=post.label
        odx['text'] = post.text if not post.is_boost else 'RT'
        odx['node_type']='post'
        odx['scores'] = post.scores
        odx['score'] = post.scores.get(SCORE_TYPE, np.nan) + (random.random()/1000)
        odx['timestamp'] = post.timestamp + (random.random()/1000)
        odx['datetime_str_h'] = post.datetime_str_h
        # odx['fixed']=dict(x=False,y=False)
        odx['num_replies'] = post.num_replies
        odx['is_read']=post.is_read
        odx['is_reply']=post.is_reply
        odx['is_boost']=post.is_boost
        return odx



